# -*- coding: utf-8 -*-
"""
PyMzn provides functions that mimic and enhance the tools from the libminizinc
library. With these tools, it is possible to compile a MiniZinc model into
FlatZinc, solve a given problem and get the output solutions directly into the
python code.

The main function that PyMzn provides is the ``minizinc`` function, which
executes the entire workflow for solving a CSP problem encoded in MiniZinc.
Solving a MiniZinc problem with PyMzn is as simple as:::

    import pymzn
    pymzn.minizinc('test.mzn')

The ``minizinc`` function is probably the way to go for most of the problems,
but the ``mzn2fzn`` and ``solns2out`` functions are in the public API to allow
for maximum flexibility. The latter two functions are wrappers of the two
homonym MiniZinc tools for, respectively, converting a MiniZinc model into a
FlatZinc one and getting custom output from the solution stream of a solver.
"""

import os
import itertools
import contextlib

from subprocess import CalledProcessError
from tempfile import NamedTemporaryFile

import pymzn.config as config

from pymzn.bin import run
from . import _solvers
from ._solvers import gecode
from ._model import MiniZincModel
from pymzn._utils import get_logger
from pymzn._dzn import dzn_eval, dzn


def minizinc(mzn, *dzn_files, data=None, keep=False, output_base=None,
             globals_dir=None, stdlib_dir=None, path=None, eval_output=True,
             output_vars=None, solver=gecode, check_complete=False,
             all_solutions=False, parse_output=True, **solver_args):
    """Implements the workflow to solve a CSP problem encoded with MiniZinc.

    It first calls mzn2fzn to compile the fzn and ozn files, then it calls the
    provided solver and in the end it calls the solns2out utility on the
    output of the solver.

    Parameters
    ----------
    mzn : str or MiniZincModel
        The minizinc problem to be solved.  It can be either a string or an
        instance of MiniZincModel.  If it is a string, it can be either the path
        to the mzn file or the content of the model.
    *dzn_files
        A list of paths to dzn files to attach to the mzn2fzn execution,
        provided as positional arguments; by default no data file is attached.
        Data files are meant to be used when there is data that is static across
        several minizinc executions.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the mzn2fnz function. The dictionary is then automatically converted to
        dzn format by the pymzn.dzn function. This property is meant to include
        data that dynamically changes across several minizinc executions.
    keep : bool
        Whether to keep the generated mzn, dzn, fzn and ozn files or not. If
        False, the generated files are created as temporary files which will be
        deleted right after the problem is solved. Though pymzn generated files
        are not originally intended to be kept, this property can be used for
        debugging purpose. Notice that in case of error the files are not
        deleted even if this parameter is False.  Default is False.
    output_base : str
        If ``keep=True``, this parameter is used as the base name for the output
        mzn, dzn, fzn and ozn files (extension are attached automatically). This
        name should include the parent directories if different from the working
        one.  Parent directories are not created automatically so they are
        required to exist. If None is provided (default) the name of the input
        file is used. If the mzn input was a content string, then the default
        name 'pymzn' and the working directory are used.
    globals_dir : str
        The name of the directory where to search for global included files in
        the standard library; by default the solver specific global library is
        used.
    stdlib_dir : str
        The name of the directory containing the standard library of minizinc.
        When None (default) the MiniZinc default is used.
    path : str or list
        One or more additional paths to search for included mzn files when
        running ``mzn2fzn``.
    eval_output : bool
        Whether to evaluate the output of the solver or solns2out function into
        Python objects with the ``pymzn.dzn_eval`` function. If the output is
        not evaluated, then a list of strings containing the solutions is
        returned. This is useful when a custom output statement is given in the
        mzn file. The default value is True. When this argument
    output_vars : list of str
        The list of output variables. If not provided, the default list is the
        list of free variables in the model, i.e. those variables that are
        declared but not defined in the model.  This argument is only used when
        ``eval_output`` is True.
    solver : Solver
        An instance of Solver to use to solve the minizinc problem. The default
        is pymzn.gecode.
    check_complete : bool
        If True, a boolean value is returned, in addition to
        the solutions of the problem, indicating the completion status of the
        problem. The returned boolean is True when the solver completed its work
        and, in case ``all_solutions=True``, returned all the solutions; it is
        False when the solver did not complete its work because e.g. was
        interrupted by a timeout.
    all_solutions : bool
        Whether all the solutions must be returned. Notice that this is only
        used if the solver supports returning all solutions, otherwise it is
        ignored. Default is False.
    **solver_args
        Additional arguments to pass to the solver, provided as additional
        keyword arguments to this function check the solver documentation for
        the available arguments.

    Returns
    -------
    list or tuple
        Returns a list of solutions. If eval_output is True, the solutions are
        returned as dictionaries of variable assignments, otherwise they are
        solution strings as returned from the solns2out function. If
        ``check_complete=True`` the result is a tuple containing the solution
        list as first argument and a boolean value indicating the completion
        status of the problem as second argument.
    """
    log = get_logger(__name__)

    if isinstance(mzn, MiniZincModel):
        mzn_model = mzn
    else:
        mzn_model = MiniZincModel(mzn)

    if eval_output and parse_output:
        mzn_model.dzn_output_stmt(output_vars)

    if not solver:
        solver = gecode
    elif isinstance(solver, str):
        solver = getattr(_solvers, solver)

    _globals_dir = globals_dir or solver.globals_dir

    output_dir = None
    output_prefix = 'pymzn'
    if keep:
        if output_base:
            output_dir, output_prefix = os.path.split(output_base)
        elif mzn_model.mzn_file:
            output_dir, mzn_name = os.path.split(mzn_model.mzn_file)
            output_prefix, mzn_ext = os.path.split(mzn_name)
        else:
            output_dir = os.getcwd()
    output_prefix += '_'
    output_file = NamedTemporaryFile(dir=output_dir, prefix=output_prefix,
                                     suffix='.mzn', delete=False, mode='w+',
                                     buffering=1)
    mzn_model.compile(output_file)
    mzn_file = output_file.name

    fzn_file, ozn_file = mzn2fzn(mzn_file, *dzn_files, data=data,
                                 keep_data=keep, path=path,
                                 globals_dir=_globals_dir,
                                 stdlib_dir=stdlib_dir)

    if not solver.support_all and all_solutions:
        log.warning('Solver does not support returning all solutions.')

    solver_check_complete = not solver.support_ozn and check_complete

    out = solver.solve(fzn_file, check_complete=solver_check_complete,
                       all_solutions=all_solutions, **solver_args)

    if solver_check_complete:
        out, complete = out

    if solver.support_ozn:
        try:
            if check_complete:
                solns, complete = solns2out(out, ozn_file, check_complete=True,
                                            parse_output=parse_output)
            else:
                solns = solns2out(out, ozn_file, parse_output=parse_output)
        except (MiniZincUnsatisfiableError, MiniZincUnknownError,
                MiniZincUnboundedError) as err:
            err.mzn_file = mzn_file
            raise err
    else:
        solns = out

    if eval_output:
        solns = list(map(dzn_eval, solns))

    if not keep:
        with contextlib.suppress(FileNotFoundError):
            if mzn_file:
                os.remove(mzn_file)
                log.debug('Deleting file: {}', mzn_file)
            if fzn_file:
                os.remove(fzn_file)
                log.debug('Deleting file: {}', fzn_file)
            if ozn_file:
                os.remove(ozn_file)
                log.debug('Deleting file: {}', ozn_file)

    if check_complete:
        return solns, complete
    return solns


def mzn2fzn(mzn_file, *dzn_files, data=None, keep_data=False, globals_dir=None,
            stdlib_dir=None, path=None, no_ozn=False):
    """Flatten a MiniZinc model into a FlatZinc one. It executes the mzn2fzn
    utility from libminizinc to produce a fzn and ozn files from a mzn one.

    Parameters
    ----------
    mzn : str or MiniZincModel
        The minizinc problem to be solved.  It can be either a string or an
        instance of MiniZincModel.  If it is a string, it can be either the path
        to the mzn file or the content of the model.
    *dzn_files
        A list of paths to dzn files to attach to the mzn2fzn execution,
        provided as positional arguments; by default no data file is attached.
        Data files are meant to be used when there is data that is static across
        several minizinc executions.
    data : dict
        Additional data as a dictionary of variables assignments to supply to
        the mzn2fnz function. The dictionary is then automatically converted to
        dzn format by the pymzn.dzn function. This property is meant to include
        data that dynamically changes across several minizinc executions.
    keep_data : bool
        Whether to write the dzn inline data provided in the ``data`` parameter
        into a file and keep it. Default is False.
    globals_dir : str
        The name of the directory where to search for global included files in
        the standard library; by default the solver specific global library is
        used.
    stdlib_dir : str
        The name of the directory containing the standard library of minizinc.
        When None (default) the MiniZinc default is used.
    path : str or list
        One or more additional paths to search for included mzn files when
        running ``mzn2fzn``.
    no_ozn : bool
        If True, the ozn file is not produced, False otherwise.

    Returns
    -------
    tuple (str, str)
        The paths to the generated fzn and ozn files. If ``no_ozn=True``, the
        second argument is None.
    """
    log = get_logger(__name__)

    args = [config.get('mzn2fzn', 'mzn2fzn')]
    if stdlib_dir:
        args.append('--stdlib-dir')
        args.append(stdlib_dir)
    if globals_dir:
        args.append('-G')
        args.append(globals_dir)
    if no_ozn:
        args.append('--no-output-ozn')
    if path:
        if isinstance(path, str):
            path = [path]
        elif not isinstance(path, list):
            raise TypeError('The path provided is not valid.')
        for p in path:
            args.append('-I')
            args.append(p)

    dzn_files = list(dzn_files)
    data_file = None
    if data:
        if isinstance(data, dict):
            data = dzn(data)
        elif isinstance(data, str):
            data = [data]
        elif not isinstance(data, list):
            raise TypeError('The additional data provided is not valid.')

        if keep_data or sum(map(len, data)) >= config.get('dzn_width', 70):
            mzn_base, __ = os.path.splitext(mzn_file)
            data_file = mzn_base + '_data.dzn'
            with open(data_file, 'w') as f:
                f.write('\n'.join(data))
            dzn_files.append(data_file)
            log.debug('Generated file: {}', data_file)
        else:
            data = ' '.join(data)
            args.append('-D')
            args.append(data)

    args += [mzn_file] + dzn_files

    try:
        run(args)
    except CalledProcessError as err:
        log.exception(err.stderr)
        raise RuntimeError(err.stderr) from err

    if not keep_data:
        with contextlib.suppress(FileNotFoundError):
            if data_file:
                os.remove(data_file)
                log.debug('Deleting file: {}', data_file)

    mzn_base = os.path.splitext(mzn_file)[0]
    fzn_file = '.'.join([mzn_base, 'fzn'])
    fzn_file = fzn_file if os.path.isfile(fzn_file) else None
    ozn_file = '.'.join([mzn_base, 'ozn'])
    ozn_file = ozn_file if os.path.isfile(ozn_file) else None

    if fzn_file:
        log.debug('Generated file: {}', fzn_file)
    if ozn_file:
        log.debug('Generated file: {}', ozn_file)

    return fzn_file, ozn_file


def solns2out(soln_stream, ozn_file, check_complete=False, parse_output=True):
    """Wraps the solns2out utility, executes it on the solution stream, and
    then returns the output.

    Parameters
    ----------
    soln_stream : str
        The solution stream returned by the solver.
    ozn_file : str
        The ozn file path produced by the mzn2fzn function.
    check_complete : bool
        If True, a boolean value is returned, in addition to
        the solutions of the problem, indicating the completion status of the
        problem.

    Returns
    -------
    list or tuple
        Returns a list of solution strings. If ``check_complete=True`` the
        result is a tuple containing the solution list as first argument and a
        boolean value indicating the completion status of the problem as second
        argument.
    """
    log = get_logger(__name__)

    soln_sep = '----------'
    search_complete_msg = '=========='
    unsat_msg = '=====UNSATISFIABLE====='
    unkn_msg = '=====UNKNOWN====='
    unbnd_msg = '=====UNBOUNDED====='

    if parse_output:
        args = [config.get('solns2out', 'solns2out'), ozn_file]
        try:
            process = run(args, stdin=soln_stream)
            out = process.stdout
        except CalledProcessError as err:
            log.exception(err.stderr)
            raise RuntimeError(err.stderr) from err
    else:
        out = soln_stream

    lines = out.splitlines()
    solns = []
    curr_out = []
    complete = False
    for line in lines:
        line = line.strip()
        if line == soln_sep:
            soln = '\n'.join(curr_out)
            solns.append(soln)
            curr_out = []
        elif line == search_complete_msg:
            complete = True
            break
        elif line == unkn_msg:
            raise MiniZincUnknownError()
        elif line == unsat_msg:
            raise MiniZincUnsatisfiableError()
        elif line == unbnd_msg:
            raise MiniZincUnboundedError()
        else:
            curr_out.append(line)

    log.debug('Solutions found: {}', len(solns))

    if check_complete:
        return solns, complete
    return solns


class MiniZincError(RuntimeError):
    """Generic error for the MiniZinc functions."""

    def __init__(self, msg=None):
        super().__init__(msg)
        self._mzn_file = None

    @property
    def mzn_file(self):
        """str: the mzn file that generated the error."""
        return self._mzn_file

    @mzn_file.setter
    def mzn_file(self, _mzn_file):
        self._mzn_file = _mzn_file


class MiniZincUnsatisfiableError(MiniZincError):
    """Error raised when a minizinc problem is found to be unsatisfiable."""

    def __init__(self):
        super().__init__('The problem is unsatisfiable.')


class MiniZincUnknownError(MiniZincError):
    """Error raised when minizinc returns no solution (unknown)."""

    def __init__(self):
        super().__init__('The solution of the problem is unknown.')


class MiniZincUnboundedError(MiniZincError):
    """Error raised when a minizinc problem is found to be unbounded."""

    def __init__(self):
        super().__init__('The problem is unbounded.')





import os
import time
import signal
import logging
import subprocess

from subprocess import Popen, PIPE, TimeoutExpired, CalledProcessError


class Process:

    def __init__(self, args):
        self.args = args
        self._process = None
        self.start_time = None
        self.end_time = None
        self.timeout = None
        self.stdout_data = None
        self.stderr_data = None
        self.expired = False
        self.completed = False
        self.async = False

    @property
    def running_time(self):
        end = self.end_time or time.time()
        return end - self.start_time

    @property
    def stdout(self):
        return self._process.stdout

    def _check_alive(self):
        self.returncode = retcode = self._process.poll()
        if retcode is None:
            return
        self.end_time = time.time()
        if retcode:
            if self._process.stderr.closed:
                stderr = self.stderr_data
            else:
                self.stderr_data = stderr = self._process.stderr.read()
            raise CalledProcessError(retcode, self.args, None, stderr)
        self.complete = True

    def _check_timeout(self):
        if not self.timeout:
            return
        if self.running_time > self.timeout:
            self.end_time = time.time()
            self.stderr_data = stderr = self._process.stderr.read()
            self.kill()
            self.expired = True
            raise TimeoutExpired(self.args, self.timeout, None, stderr)

    def kill(self):
        if not self._process:
            raise RuntimeError('The process has not started yet')
        self._process.terminate()
        self._process.kill()
        self._process.wait()
        self._process.stdout.close()
        self._process.stderr.close()

    def lines(self):
        if not self._process:
            raise RuntimeError('The process has not started yet')
        stdout = self._process.stdout
        while not stdout.closed:
            self._check_alive()
            self._check_timeout()
            line = stdout.readline()
            if line == '':
                break
            yield line

    def __iter__(self):
        return self.lines()

    def run_async(self, stdin=None, timeout=None):
        if self._process:
            raise RuntimeError('Process already started')
        self.timeout = timeout
        self.async = True
        self.start_time = time.time()
        self._process = subprocess.Popen(
            self.args, bufsize=0, universal_newlines=True, stdin=stdin,
            stdout=PIPE, stderr=PIPE
        )
        return self

    def run(self, input=None, timeout=None):
        if self._process:
            raise RuntimeError('Process already started')

        self.start_time = time.time()
        self.timeout = timeout
        with subprocess.Popen(
                self.args, bufsize=1, universal_newlines=True,
                stdin=PIPE, stdout=PIPE, stderr=PIPE
        ) as self._process:
            try:
                self.stdout_data, self.stderr_data = \
                        self._process.communicate(input, timeout)
                self._check_alive()
            except KeyboardInterrupt:
                self.kill()
                self.returncode = 0
            except TimeoutExpired:
                self.stdout_data, self.stderr_data = self._process.communicate()
                self._check_alive()
                self.kill()
                raise TimeoutExpired(
                    self.args, self.timeout, self.stdout_data, self.stderr_data
                )
            finally:
                self.end_time = time.time()
        return self

    def close(self):
        self.kill()
        self._check_alive()




# -*- coding: utf-8 -*-
"""Provides classes to interface solvers with PyMzn.

PyMzn interfaces with solvers through the ``Solver`` class. This class includes
the necessary infomation for PyMzn to setup the solver. This class also include
the ``solve`` method which takes care of the actual solving of a FlatZinc model.
The solvers classes are subclasses of the ``Solver`` class, providing
implementations of the ``solve`` method.
PyMzn provides a number of solver implementations out-of-the-box.
PyMzn's default solver is Gecode, which class is `pymzn.Gecode` and the default
instance is ``pymzn.gecode``.

To use a different solver or to exend an existing one, one has to subclass the
Solver class and implement the ``solve`` method.

For instance:::

    from pymzn.bin import run
    from pymzn import Solver

    class MySolver(Solver):
        def __init__(self, path='/path/to/solver'):
            self.path = path
            support_ozn = True
            support_all = False
            super().__init__(support_ozn, support_all)

        def solve(self, fzn_file, *, all_solutions=False, check_complete=False,
                  arg1=def_val1, arg2=def_val2, **kwargs):
        args = [self.path, '-arg1', arg1, '-arg2', arg2, fzn_file]
        process = run(args)
        return process.stdout


Then one can run the ``minizinc`` function with the custom solver:::

    pymzn.minizinc('test.mzn', solver=MySolver(), arg1=val1, arg2=val2)
"""

import pymzn.config as config

from pymzn.bin import run
from pymzn._utils import get_logger

from subprocess import CalledProcessError


class Solver(object):
    """Abstract solver class.

    All solver classes inherit from this class and provide implementations for
    the ``solve`` method.

    Attributes
    ----------
    support_ozn : bool
        Whether the solver's output can be parsed by ``solns2out``.
    support_all : bool
        Whether the solver supports the output of all solutions.
    globals_dir : str
        The directory containing the solver-specific redefinitions of global
        constraints, used when calling ``mzn2fzn``. This should be the default
        for the solver. When None is provided (default) the ``std`` directory is
        used.
    """
    def __init__(self, support_ozn, support_all, globals_dir=None):
        self.support_ozn = support_ozn
        self.support_all = support_all
        self.globals_dir = globals_dir

    def solve(self, fzn_file, *, all_solutions=False,
              check_complete=False, **kwargs):
        """Solve a problem encoded in FlatZinc.

        This method should call an external solver, wait for the solution and
        provide the output of the solver. If the solver does not have a Python
        interface, the ``pymzn.bin`` module can be used to run external
        executables.

        If ``support_ozn`` is ``False``, the solver should directly provide
        solutions compliant to the PyMzn solution format. This means that the
        the implementation of this method should parse the output of the solver
        and return the solutions as dictionaries of variable assignments (see
        the ``eval_dzn`` method). In alternative the solutions should be
        returned in dzn format, so they can be evaluated by the ``eval_dzn``
        function.

        Parameters
        ----------
        fzn_file : str
            The path to the fzn file to use as input of the solver.
        all_solutions : bool
            Whether the solver should output all the solutions. If
            ``support_all=False`` then the implementation should ignore the
            value of this parameter.
        check_complete : bool
            Whether the solver should return a second boolean value indicating
            if the search was completed successfully.
        **kwargs
            Additional arguments for the solver provided through the
            ``pymzn.minizinc`` function.

        Returns
        -------
        str, list or tuple
            This method should return a string if the solver supports ozn
            parsing via ``solns2out``. The string simply contains the output of
            the solver.
            If the solver does not support the ozn parsing, then it should parse
            the solvers output by itself and thus return a list of solutions in
            PyMzn format (dictionaries of variable assignments, as in
            ``pymzn.eval_dzn``).

            In both cases, if ``check_complete=True`` then the method should
            return a tuple containing the above output and a second boolean
            indicating whether the search was completed successfully.
        """
        raise NotImplementedError()


class Gecode(Solver):
    """Interface to the Gecode MILP solver.

    Parameters
    ----------
    path : str
        The path to the Gecode executable. If None, ``fzn-gecode`` is used.
    """
    def __init__(self, path=None):
        super().__init__(True, True, globals_dir='gecode')
        self.cmd = path or 'fzn-gecode'

    def solve(self, fzn_file, *, check_complete=False, all_solutions=False,
              timeout=0, parallel=1, n_solns=-1, seed=0, restart=None,
              restart_base=None, restart_scale=None, suppress_segfault=False,
              **kwargs):
        """Solves a problem with the Gecode solver.

        Parameters
        ----------
        fzn_file : str
            The path to the fzn file to use as input of the solver.
        all_solutions : bool
            Whether the solver should output all the solutions. Equivalent to
            ``n_solns=0``.
        check_complete : bool
            Whether the solver should return a second boolean value indicating
            if the search was completed successfully.
        n_solns : int
            The number of solutions to output (0 = all, -1 = one/best);
            the default is -1.
        parallel : int
            The number of threads to use to solve the problem
            (0 = #processing units); default is 1.
        time : int or float
            The time cutoff in seconds, after which the execution is truncated
            and the best solution so far is returned, 0 means no time cutoff;
            default is 0.
        seed : int
            Random seed; default is 0.
        restart : str
            Restart sequence type; default is None.
        restart_base : str
            Base for geometric restart sequence; if None (default) the default
            value of Gecode is used, which is 1.5.
        restart_scale : str
            Scale factor for restart sequence; if None (default) the default
            value of Gecode is used, which is 250.
        suppress_segfault : bool
            Whether to accept or not a solution returned when a segmentation
            fault has happened (this is unfortunately necessary sometimes due to
            some bugs in Gecode).

        Returns
        -------
        str or tuple
            A string containing the solution output stream of the execution of
            Gecode on the specified problem; it can be directly be given to the
            function solns2out to be evaluated. If ``check_complete=True``
            returns an additional boolean, checking whether the search was
            completed before the timeout.
        """
        log = get_logger(__name__)

        args = [self.cmd]
        if n_solns > 0:
            args.append('-n')
            args.append(str(n_solns))
        elif n_solns == 0 or all_solutions:
            args.append('-a')
        if parallel != 1:
            args.append('-p')
            args.append(str(parallel))
        if timeout > 0:
            args.append('-time')
            args.append(str(timeout * 1000)) # Gecode takes milliseconds
        if seed != 0:
            args.append('-r')
            args.append(str(seed))
        if restart:
            args.append('-restart')
            args.append(str(restart))
        if restart_base:
            args.append('-restart-base')
            args.append(str(restart_base))
        if restart_scale:
            args.append('-restart-scale')
            args.append(str(restart_scale))
        args.append(fzn_file)

        try:
            process = run(args)
            if timeout > 0:
                process.timeout = timeout
                if process.time >= timeout:
                    complete = False
            out = process.stdout
        except CalledProcessError as err:
            if suppress_segfault:
                log.warning('Gecode returned error code {} (segmentation '
                            'fault) but a solution was found and returned '
                            '(suppress_segfault=True).'.format(err.returncode))
                out = err.stdout
                complete = False
            else:
                log.exception(err.stderr)
                raise RuntimeError(err.stderr) from err
        if check_complete:
            return out, complete
        return out


class Optimathsat(Solver):
    """Simple wrapper for the OptiMathSat solver.

    This is a simple interface to OptiMathSat which only specifies the input
    format as a FlatZinc model, without providing any additional arguments.

    Parameters
    ----------
    path : str
        The path to the OptiMathSat executable. If None, ``optimathsat`` is used.
    """
    def __init__(self, path=None):
        super().__init__(False, False, globals_dir='std')
        self.cmd = path or 'optimathsat'

    def solve(self, fzn_file, *, all_solutions=False, check_complete=False,
              **kwargs):
        """Solves a problem with the OptiMathSat solver.

        Parameters
        ----------
        fzn_file : str
            The path to the fzn file to use as input of the solver.
        all_solutions : bool
            Ignored.
        check_complete : bool
            Ignored.
        """
        args = [self.cmd, '-input=fzn', fzn_file]

        log = get_logger(__name__)
        try:
            process = run(args)
            out = process.stdout
        except CalledProcessError as err:
            log.exception(err.stderr)
            raise RuntimeError(err.stderr) from err

        if check_complete:
            return out, True
        return out


class Opturion(Solver):
    """Interface to the Opturion CPX solver.

    Parameters
    ----------
    path : str
        The path to the Opturion executable. If None, ``fzn-cpx`` is used.
    """

    def __init__(self, path=None):
        super().__init__(True, True, globals_dir='opturion-cpx')
        self.cmd = path or 'fzn-cpx'

    def solve(self, fzn_file, *, all_solutions=False, check_complete=False,
              timeout=None, **kwargs):
        """Solves a problem with the Opturion CPX solver.

        Parameters
        ----------
        fzn_file : str
            The path to the fzn file to use as input of the solver.
        all_solutions : bool
            Whether to return all solutions.
        check_complete : bool
            Whether to return a second boolean value indicating if the search
            was completed successfully.
        timeout : int or float
            The time cutoff in seconds, after which the execution is truncated
            and the best solution so far is returned, 0 or None means no cutoff;
            default is None.

        Returns
        -------
        str or tuple
            A string containing the solution output stream of the execution of
            Opturion on the specified problem; it can be directly be given to
            the function solns2out to be evaluated. If ``check_complete=True``
            returns an additional boolean, checking whether the search was
            completed before the timeout.
        """
        args = [self.cmd]

        if timeout or all_solutions:
            args.append('-a')

        args.append(fzn_file)

        log = get_logger(__name__)

        try:
            process = run(args, timeout=timeout)
            complete = not process.expired
            out = process.stdout
        except CalledProcessError as err:
            log.exception(err.stderr)
            raise RuntimeError(err.stderr) from err

        if check_complete:
            return out, complete
        return out


class Gurobi(Solver):
    """Interface to the Gurobi solver.

    Parameters
    ----------
    path : str
        The path to the Gurobi executable. If None, ``mzn-gurobi`` is used.
    """

    def __init__(self, path=None):
        super().__init__(True, True, globals_dir='std')
        self.cmd = path or 'mzn-gurobi'

    def solve(self, fzn_file, *, all_solutions=False, check_complete=False,
              timeout=None, **kwargs):
        """Solves a problem with the Opturion Gurobi solver.

        Parameters
        ----------
        fzn_file : str
            The path to the fzn file to use as input of the solver.
        all_solutions : bool
            Whether to return all solutions.
        check_complete : bool
            Whether to return a second boolean value indicating if the search
            was completed successfully.
        timeout : int or float
            The time cutoff in seconds, after which the execution is truncated
            and the best solution so far is returned, 0 or None means no cutoff;
            default is None.

        Returns
        -------
        str or tuple
            A string containing the solution output stream of the execution of
            Opturion on the specified problem; it can be directly be given to
            the function solns2out to be evaluated. If ``check_complete=True``
            returns an additional boolean, checking whether the search was
            completed before the timeout.
        """
        args = [self.cmd]

        if timeout or all_solutions:
            args.append('-a')

        args.append(fzn_file)

        log = get_logger(__name__)

        try:
            process = run(args, timeout=timeout)
            complete = not process.expired
            out = process.stdout
        except CalledProcessError as err:
            log.exception(err.stderr)
            raise RuntimeError(err.stderr) from err

        if check_complete:
            return out, complete
        return out


#: Default Gecode instance.
gecode = Gecode(path=config.get('gecode'))

#: Default Optimathsat instance.
optimathsat = Optimathsat(path=config.get('optimathsat'))

#: Default Opturion instance.
opturion = Opturion(path=config.get('opturion'))

#: Default Gurobi instance.
gurobi = Opturion(path=config.get('gurobi'))




# -*- coding: utf-8 -*-

from ._parse import *
from ._model import *
from ._solvers import *
from ._minizinc import *

__all__ = ['minizinc', 'mzn2fzn', 'solns2out', 'MiniZincError',
           'MiniZincUnsatisfiableError', 'MiniZincUnknownError',
           'MiniZincUnboundedError', 'MiniZincModel', 'Statement', 'Constraint',
           'Variable', 'ArrayVariable', 'OutputStatement', 'SolveStatement',
           'Solver', 'Gecode', 'Optimathsat', 'Opturion', 'gecode',
           'optimathsat', 'opturion', 'parse']



# -*- coding: utf-8 -*-
"""Convenience module for running executables.

PyMzn executes the MiniZinc tools and the solvers with the functions
contained in the standard ``subprocess`` module. As a convenience method, PyMzn
provides the method ``pymzn.bin.run`` which takes the arguments of a binary
executable, as for the ``subprocess.Popen`` constructor. This method returns an
instance of ``pymzn.bin.TimedCompletedProcess``, a subclass of
``subprocess.CompletedProcess`` which contains also information about the
timeout.
"""
import os
import time
import signal
import subprocess

from ._utils import get_logger


class TimedCompletedProcess(subprocess.CompletedProcess):
    """Contains information about a completed process.

    An instance of this class is always returned by the ``pymzn.bin.run``
    method. This is a subclass of ``subprocess.CompletedProcess`` including
    additional information about running time and process timeout.

    Attributes
    ----------
    args : list or str
        The args passed to run().
    returncode : int
        The exit code of the process, negative for signals.
    time : float
        The running time in seconds.
    timeout : int or float
        The timeout given to the process.
    expired : bool
        Whether the time has expired and the process has been terminated before
        finishing.
    stdout : bytes
        The standard output (None if not captured).
    stderr : bytes
        The standard error (None if not captured).
    """
    def __init__(self, args, returncode, time, timeout=None, stdout=None,
                 stderr=None):
        super().__init__(args, returncode, stdout=stdout, stderr=stderr)
        self.time = time
        self.timeout = timeout
        self.expired = timeout and time >= timeout

    def __repr__(self):
        args = ['args={!r}'.format(self.args),
                'returncode={!r}'.format(self.returncode),
                'time={!r}'.format(self.time),
                'timeout={!r}'.format(self.timeout)]
        if self.stdout is not None:
            args.append('stdout={!r}'.format(self.stdout))
        if self.stderr is not None:
            args.append('stderr={!r}'.format(self.stderr))
        return "{}({})".format(type(self).__name__, ', '.join(args))

    def check_returncode(self):
        """Raise CalledProcessError if the exit code is non-zero and the time
        has not expired.
        """
        if not self.expired:
            super().check_returncode()


def run(args, stdin=None, timeout=None):
    """Executes a command and waits for the result.

    Parameters
    ----------
    args : list
        The list of arguments for the program to execute. Arguments should be
        formatted as for the ``subprocess.Popen`` constructor.
    stdin : str or bytes
        String or bytes containing the input stream for the process.
    timeout : int or float
        The timeout for the process in seconds.

    Returns
    -------
    TimedCompletedProcess
        An instance of TimedCompletedProcess containing the information about
        the executed process, including stdout, stderr and running time.

    Raises
    ------
    CalledProcessError
        When the process returns an error.
    """
    log = get_logger(__name__)
    log.debug('Executing command with args: {}', args)
    start = time.time()
    with subprocess.Popen(args, bufsize=1, universal_newlines=True,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          # preexec_fn not supported on windows, find a better
                          # solution, maybe this all works even without the
                          # line below
                          preexec_fn=os.setsid) as process:
        try:
            out, err = process.communicate(stdin, timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            os.killpg(os.getpgid(process.pid), signal.SIGTERM)
            out, err = process.communicate()
        ret = process.poll()
        end = time.time()
    elapsed = end - start
    log.debug('Done. Running time: {0:.2f} seconds'.format(elapsed))
    process = TimedCompletedProcess(args, ret, elapsed, timeout, out, err)
    process.check_returncode()
    return process





import re

stmt_p = re.compile('(?:^|;)\s*([^;]+)')
block_comm_p = re.compile('/\*.*\*/', re.DOTALL)
line_comm_p = re.compile('%.*\n')
var_p = re.compile('\s*([\s\w,\.\(\)\[\]\{\}\+\-\*/]+?):\s*(\w+)\s*(?:=\s*(.+))?\s*')
type_p = re.compile('\s*(?:int|float|set\s+of\s+[\s\w\.]+|array\[[\s\w\.]+\]\s*of\s*[\s\w\.]+)\s*')
var_type_p = re.compile('\s*.*?var.+\s*')
array_type_p = re.compile('\s*array\[([\s\w\.]+(?:\s*,\s*[\s\w\.]+)*)\]\s+of\s+(.+)\s*')
output_stmt_p = re.compile('\s*output\s*\[(.+?)\]\s*(?:;)?\s*')
solve_stmt_p = re.compile('\s*solve\s*([^;]+)\s*(?:;)?\s*')
constraint_p = re.compile('\s*constraint\s*(.+)\s*', re.DOTALL)


def parse(model):
    model = block_comm_p.sub('', model)
    model = line_comm_p.sub('', model)
    stmts = stmt_p.findall(model)
    parameters = []
    variables = []
    constraints = []
    solve = None
    output = None
    others = []
    for stmt in stmts:
        if not stmt.strip():
            continue
        var_m = var_p.match(stmt)
        if var_m and not ('function' in stmt or 'predicate' in stmt):
            vartype = var_m.group(1)
            name = var_m.group(2)
            value = var_m.group(3)
            var_type_m = var_type_p.match(vartype)
            if not var_type_m:
                par = name, vartype, value
                parameters.append(par)
                continue
            var = name, vartype, value
            variables.append(var)
            continue
        constraint_m = constraint_p.match(stmt)
        if constraint_m:
            constraint = constraint_m.group(1)
            constraints.append(constraint)
            continue
        solve_stmt_m = solve_stmt_p.match(stmt)
        if solve_stmt_m:
            solve = solve_stmt_m.group(1)
            continue
        output_stmt_m = output_stmt_p.match(stmt)
        if output_stmt_m:
            output = output_stmt_m.group(1)
            continue
        others.append(stmt)
    return parameters, variables, constraints, solve, output, others




# -*- coding: utf-8 -*-
"""PyMzn can also be used to dynamically change a model during runtime. For
example, it can be useful to add constraints incrementally or change the solving
statement dynamically. To dynamically modify a model, you can use the class
``MiniZincModel``, which can take an optional model file as input which can then
be modified by adding variables and constraints, and by modifying the solve or
output statements. An instance of ``MiniZincModel`` can then be passed directly
to the ``minizinc`` function to be solved.
::

    model = pymzn.MiniZinModel('test.mzn')

    for i in range(10):
        model.add_constraint('arr_1[i] < arr_2[i]')
        pymzn.minizinc(model)
"""

import re
import os.path

from pymzn._utils import get_logger
from pymzn._dzn._marsh import dzn_statement, dzn_value
from pymzn._mzn._parse import *


class Statement(object):
    """A statement of a MiniZincModel.

    Attributes
    ----------
    stmt : str
        The statement string.
    """
    def __init__(self, stmt):
        self.stmt = stmt

    def __str__(self):
        return self.stmt


class Comment(Statement):
    """A comment statement.

    Attributes
    ----------
    comment : str
        The comment string.
    """
    def __init__(self, comment):
        self.comment = comment
        stmt = '% {}\n'.format(comment)
        super().__init__(stmt)


class Constraint(Statement):
    """A constraint statement.

    Attributes
    ----------
    constr : str
        The content of the constraint, i.e. only the actual constraint without
        the starting 'constraint' and the ending semicolon.
    """
    def __init__(self, constr, comment=None):
        self.constr = constr
        stmt = 'constraint {};'.format(constr)
        super().__init__(stmt)


class Parameter(Statement):
    """A parameter statement.

    Attributes
    ----------
    par : (str, str) or (str, obj)
        Either a (name, type) pair or a (name, value) pair. In the latter case
        the type is automatically inferred from the value.
    assign : bool
        If True the parameter value will be assigned directly into the model,
        otherwise it will only be declared in the model and then it will have to
        be assigned in the data.
    """
    def __init__(self, *par, assign=True):
        name, par = par
        self.name = name
        self.type = None
        self.value = None
        self.assign = assign
        if isinstance(par, str):
            type_m = type_p.match(par)
            if type_m:
                self.type = par
        if not self.type:
            self.value = par
        if self.type:
            stmt = '{}: {};'.format(self.type, self.name)
        else:
            stmt = dzn_statement(self.name, self.value, assign=assign)
        super().__init__(stmt)


class Variable(Statement):
    """A variable statement.

    Attributes
    ----------
    name : str
        The name of the variable.
    vartype : str
        The type of the variable.
    value : str
        The optional value of the variable statement.
    output : bool
        Whether the variable is an output variable.
    """
    def __init__(self, name, vartype, value=None, output=False):
        self.name = name
        self.value = value
        self.output = output

        array_type_m = array_type_p.match(vartype)
        if array_type_m:
            indexset = array_type_m.group(1)
            domain = array_type_m.group(2)
            if 'var' not in domain:
                vartype = 'array[{}] of var {}'.format(indexset, domain)
        elif 'var' not in vartype:
            vartype = 'var ' + vartype
        self.vartype = vartype

        stmt = '{}: {}'.format(vartype, name)
        if output:
            if array_type_m:
                stmt += ' :: output_array([{}])'.format(indexset)
            else:
                stmt += ' :: output_var'
        if value:
            stmt += ' = {}'.format(value)
        stmt += ';'

        super().__init__(stmt)


class ArrayVariable(Variable):
    """An array variable statement.

    Attributes
    ----------
    name : str
        The name of the variable.
    indexset : str
        The indexset of the array.
    domain : str
        The domain of the array.
    value : str
        The optional value of the variable statement.
    output : bool
        Whether the array variable is an output array.
    """
    def __init__(self, name, indexset, domain, value=None, output=False):
        self.indexset = indexset
        self.domain = domain
        vartype = 'array[{}] of var {}'.format(indexset, domain)
        super().__init__(name, vartype, value, output)


class OutputStatement(Statement):
    """An output statement.

    Attributes
    ----------
    output : str
        The content of the output statement, i.e. only the actual output without
        the starting 'output', the square brackets and the ending semicolon.
    """
    def __init__(self, output):
        self.output = output
        stmt = 'output [{}];'.format(output)
        super().__init__(stmt)


class SolveStatement(Statement):
    """A solve statement.

    Attributes
    ----------
    solve : str
        The content of the solve statement, i.e. only the actual solve without
        the starting 'solve' and the ending semicolon.
    """

    def __init__(self, solve):
        self.solve = solve
        stmt = 'solve {};'.format(solve)
        super().__init__(stmt)


class MiniZincModel(object):
    """Mutable class representing a MiniZinc model.

    It can use a mzn file as template, add variables and constraints,
    modify the solve and output statements. The output statement can also be
    replaced by a dzn representation of a list of output variables.
    The final model is a string combining the existing model (if provided)
    and the updates performed on the MiniZincModel instance.

    Parameters
    ----------
    mzn : str
        The content or the path to the template mzn file.
    """
    def __init__(self, mzn=None):
        self._statements = []
        self._solve_stmt = None
        self._output_stmt = None
        self._free_vars = set()
        self._array_dims = {}
        self._modified = False
        self._parsed = False

        self.mzn_file = None
        self.model = None
        if mzn and isinstance(mzn, str):
            if os.path.isfile(mzn):
                self.mzn_file = mzn
            else:
                self.model = mzn

    def comment(self, comment):
        """Add a comment to the model.

        Parameters
        ----------
        comment : str
            The comment string.
        """
        if not isinstance(comment, Comment):
            comment = Comment(comment)
        self._statements.append(comment)
        self._modified = True

    def parameter(self, *par, assign=True):
        """Adds a parameter to the model.

        Parameters
        ----------
        par : Parameter or (str, str) or (str, obj)
            Either a Parameter, a (name, type) pair or a (name, value) pair. In
            the latter case, the type is inferred automatically from the value.
        assign : bool
            If True the parameter value will be assigned directly into the
            model, otherwise it will only be declared in the model and then it
            will have to be assigned in the data.
        """
        if isinstance(par[0], Parameter):
            par = par[0]
        else:
            par = Parameter(*par, assign=assign)
        self._statements.append(par)
        self._modified = True

    def parameters(self, pars, assign=True):
        """Add a list of parameters.

        Parameters
        ----------
        pars : list of (name, val) or Parameter
            The list of parameters to add to the model.
        assign : bool
            If True the parameters value will be assigned directly into the
            model, otherwise they will only be declared in the model and then
            they will have to be assigned in the data.
        """
        for par in pars:
            self.parameter(*par, assign=assign)
        self._modified = True

    def variable(self, name, vartype, value=None, output=False):
        """Adds a variable to the model.

        Parameters
        ----------
        name : str
            The name of the variable.
        vartype : str
            The type of the variable.
        value : str
            The optional value of the variable statement.
        output : bool
            Whether the variable is an output variable.
        """
        value = dzn_value(value) if value is not None else None
        array_type_m = array_type_p.match(vartype)
        if array_type_m:
            indexset = array_type_m.group(1)
            domain = array_type_m.group(2)
            dim = len(indexset.split(','))
            self._array_dims[name] = dim
            var = ArrayVariable(name, indexset, domain, value, output)
        else:
            var = Variable(name, vartype, value, output)
        self._statements.append(var)

        var_type_m = var_type_p.match(vartype)
        if output or var_type_m and value is None:
            self._free_vars.add(name)
        self._modified = True

    def array_variable(self, name, indexset, domain, value=None, output=False):
        """Adds an array variable to the model.

        Parameters
        ----------
        name : str
            The name of the array.
        indexset : str
            The indexset of the array.
        domain : str
            The domain of the array.
        value : str
            The optional value of the array variable statement.
        output : bool
            Whether the array variable is an output array.
        """
        value = dzn_value(value) if value is not None else None
        var = ArrayVariable(name, indexset, domain, value, output)
        self._statements.append(var)
        if output or var_type_p.match(domain) and value is None:
            self._free_vars.add(name)
        self._array_dims[var] = len(indexset.split(','))
        self._modified = True

    def constraint(self, constr):
        """Adds a constraint to the current model.

        Parameters
        ----------
        constr : str or Constraint
            As a string, the content of the constraint, i.e. only the actual
            constraint without the starting 'constraint' and the ending
            semicolon.
        """
        if not isinstance(constr, Constraint):
            constr = Constraint(constr)
        self._statements.append(constr)
        self._modified = True

    def constraints(self, constrs):
        """Add a list of constraints.

        Parameters
        ----------
        constrs : list of str or Constraint
            A list of constraints.
        """
        for constr in constrs:
            self.constraint(constr)

    def solve(self, solve_stmt):
        """Updates the solve statement of the model.

        Parameters
        ----------
        solve_stmt : str
            The content of the solve statement, i.e. only the actual solve
            without the starting 'solve' and the ending semicolon.
        """
        if not isinstance(solve_stmt, SolveStatement):
            solve_stmt = SolveStatement(solve_stmt)
        self._solve_stmt = solve_stmt
        self._modified = True

    def satisfy(self):
        """Shorthand for solve('satisfy')"""
        self._solve_stmt = SolveStatement('satisfy')
        self._modified = True

    def maximize(self, expr):
        """Shorthand for solve('maximize ' + expr)"""
        self._solve_stmt = SolveStatement('maximize ' + expr)
        self._modified = True

    def minimize(self, expr):
        """Shorthand for solve('minimize ' + expr)"""
        self._solve_stmt = SolveStatement('minimize ' + expr)
        self._modified = True

    def output(self, output_stmt):
        """Updates the output statement of the model.

        Parameters
        ----------
        solve_stmt : str
            The content of the output statement, i.e. only the actual output
            without the starting 'output', the square brackets and the ending
            semicolon.
        """
        if not isinstance(output_stmt, OutputStatement):
            output_stmt = OutputStatement(output_stmt)
        self._output_stmt = output_stmt
        self._modified = True

    def _load_model(self):
        if not self.model:
            if self.mzn_file:
                with open(self.mzn_file) as f:
                    self.model = f.read()
            else:
                self.model = ''
        return self.model

    def _parse_model_stmts(self):
        if self._parsed:
            return
        model = self._load_model()
        _, variables, *_ = parse(model)
        for var in variables:
            name, vartype, value = var
            if var_type_p.match(vartype) and value is None:
                self._free_vars.add(name)
            array_type_m = array_type_p.match(vartype)
            if array_type_m:
                dim = len(array_type_m.group(1).split(','))
                self._array_dims[name] = dim
        self._parsed = True

    def dzn_output_stmt(self, output_vars=None, comment=None):
        """Sets the output statement to be a dzn representation of output_vars.

        If output_var is not provided (= None) then the free variables of the
        model are used i.e. those variables that are declared but not defined in
        the model (not depending on other variables).

        Parameters
        ----------
        output_vars : list of str
            The list of output variables.
        comment : str
            A comment to attach to the output statement.
        """

        # Look for free variables and array dimensions in the model statements
        self._parse_model_stmts()

        # Set output vars to the free variables if None provided
        if output_vars is None:
            output_vars = list(self._free_vars)

        if not output_vars:
            return

        # Build the output statement from the output variables
        out_var = '"{0} = ", show({0}), ";\\n"'
        out_array = '"{0} = array{1}d(", {2}, ", ", show({0}), ");\\n"'
        out_list = []
        for var in output_vars:
            if var in self._array_dims:
                dim = self._array_dims[var]
                if dim == 1:
                    show_idx_sets = 'show(index_set({}))'.format(var)
                else:
                    show_idx_sets = []
                    for d in range(1, dim + 1):
                        show_idx_sets.append('show(index_set_{}of{}'
                                             '({}))'.format(d, dim, var))
                    show_idx_sets = ', ", ", '.join(show_idx_sets)
                out_list.append(out_array.format(var, dim, show_idx_sets))
            else:
                out_list.append(out_var.format(var))
        out_list = ', '.join(out_list)
        self.output(out_list)

    def compile(self, output_file=None):
        """Compiles the model and writes it to file.

        The compiled model contains the content of the template (if provided)
        plus the added variables and constraints. The solve and output
        statements will be replaced if new ones are provided.

        Parameters
        ----------
        output_file : file-like
            The file where to write the compiled model.

        Returns
        -------
        str
            A string containing the generated model.
        """
        model = self._load_model()

        if self._modified:
            lines = ['\n\n\n%%% GENERATED BY PYMZN %%%\n\n']

            for stmt in self._statements:
                lines.append(str(stmt))

            if self._solve_stmt:
                model = solve_stmt_p.sub('', model)
                lines.append(str(self._solve_stmt) + '\n')

            if self._output_stmt:
                model = output_stmt_p.sub('', model)
                lines.append(str(self._output_stmt) + '\n')

            model += '\n'.join(lines)

        if output_file:
            output_file.write(model)

        return model

